require 'rest-client'
require 'nokogiri'

module PredicsisMlSdk
  class Requests
    extend Util

    MODELS = %w(source dataset dictionary preparation_rules_set modalities_set report model data_file)

    class << self
      %w(get post patch delete).each do |method|
        define_method("#{method}") do |url, params, headers = {}|
          with_captured_console do
            uri = URI(PredicsisMlSdk.config.predicsis_api_host + url)
            res = Net::HTTP.start(uri.host, uri.port) do |http|
              request = Net::HTTP.const_get(method.capitalize).new(uri, headers)
              request.body = params if params

              http.request(request)
            end

            case res
            when Net::HTTPSuccess, Net::HTTPRedirection
              return JSON.parse(res.body) unless res.body.empty?
            else
              fail RequestError.new(res['message'], res.body.empty? ? res.body : JSON.parse(res.body))
            end
          end
        end
      end

      def post_token(token, params)
        PredicsisMlSdk.logger.debug 'POST settings/tokens'
        params = { token: params }.to_json
        response = post('/settings/tokens', params, headers(token))
        PredicsisMlSdk.logger.debug "Personal Access Token: #{response['token']['token']}"
        response['token']
      end

      def get_token(token, id)
        PredicsisMlSdk.logger.debug "GET settings/tokens/#{id}"
        get("/settings/tokens/#{id}", nil, headers(token))['token']
      end

      def get_tokens(token)
        PredicsisMlSdk.logger.debug 'GET settings/tokens'
        get('/settings/tokens', nil, headers(token))['tokens']
      end

      def delete_token(token, id)
        PredicsisMlSdk.logger.debug "DELETE settings/tokens/#{id}"
        delete("/settings/tokens/#{id}", nil, headers(token))
      end

      def initiate_upload(token)
        PredicsisMlSdk.logger.debug 'INIT upload'
        post('/uploads', nil, headers(token))['upload']
      end

      def get_part_url(token, id, part_number, path)
        get("/uploads/#{id}?part_number=#{part_number}&path=#{path}", nil, headers(token))['upload']
      end

      def complete_upload(token, id, path)
        PredicsisMlSdk.logger.debug 'COMPLETE upload'
        patch("/uploads/#{id}?path=#{path}", nil, headers(token))
      end

      def upload(endpoint, path, part)
        RestClient.put(endpoint, File.open(path), { 'Content-Length' => File.size(path).to_s })
      end

      def get_job(token, id)
        get("/jobs/#{id}", nil, headers(token))['job']
      end

      def get_jobs(token)
        PredicsisMlSdk.logger.debug 'GET jobs'
        get('/jobs', nil, headers(token))['jobs']
      end

      def delete_job(token, id)
        PredicsisMlSdk.logger.debug "DELETE job/#{id}"
        delete("/jobs/#{id}", nil, headers(token))
      end

      def get_variables(token, parent_id)
        PredicsisMlSdk.logger.debug "GET dictionaries/#{parent_id}/variables"
        get("/dictionaries/#{parent_id}/variables", nil, headers(token))['variables']
      end

      def get_variable(token, id, parent_id)
        PredicsisMlSdk.logger.debug "GET dictionaries/#{parent_id}/variables/#{id}"
        get("/dictionaries/#{parent_id}/variables/#{id}", nil, headers(token))['variable']
      end

      def patch_variable(token, id, parent_id, params = {})
        PredicsisMlSdk.logger.debug "PATCH dictionaries/#{parent_id}/variables/#{id}"
        params = { variable: params }.to_json
        patch("/dictionaries/#{parent_id}/variables/#{id}",
          params, headers(token))['variable']
      end

      def get_signed_url(token, id)
        PredicsisMlSdk.logger.debug "GET data_files/#{id}/signed_url"
        get("/data_files/#{id}/signed_url", nil, headers(token))['data_file']
      end

      MODELS.each do |model|
        define_method("post_#{model}") do |token, params = {}, other_params = {}|
          PredicsisMlSdk.logger.debug "POST #{model.pluralize}"
          params = { :"#{model}" => params }.merge(other_params)
          response = post("/#{model.pluralize}", params.to_json, headers(token))
          PredicsisMlSdk.logger.info("#{model}_id: #{response[model]['id']}")
          response[model]
        end

        define_method("delete_#{model}") do |token, id|
          PredicsisMlSdk.logger.debug "DELETE #{model.pluralize}/#{id}"
          delete("/#{model.pluralize}/#{id}", nil, headers(token))
        end

        define_method("get_#{model}") do |token, id|
          PredicsisMlSdk.logger.debug "GET #{model.pluralize}/#{id}"
          get("/#{model.pluralize}/#{id}", nil, headers(token))[model]
        end

        define_method("get_#{model.pluralize}") do |token|
          PredicsisMlSdk.logger.debug "GET #{model.pluralize}"
          get("/#{model.pluralize}", nil, headers(token))[model.pluralize]
        end

        define_method("patch_#{model}") do |token, id, params = {}|
          params = { :"#{model}" => params }.to_json
          PredicsisMlSdk.logger.debug "PATCH #{model.pluralize}/#{id}"
          patch("/#{model.pluralize}/#{id}", params, headers(token))[model]
        end
      end

      private

      def headers(token)
        @headers = {
          'Accept' => 'application/json',
          'Content-Type' => 'application/json',
          'Authorization' => "Bearer #{token}"
        }
      end
    end
  end
end
